# -*- coding: utf8 -*-
from __future__ import unicode_literals

import time

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimString,
    Segment,
)


class AmsSegmentMSH(HeaderSegment):
    sender_name = HprimString((3, 1))
    sender_version = HprimString((3, 2), default="1.0.0")
    user_configured = HprimString((5, 3), default="Abbott")
    date = HprimDate(7)
    message_type = HprimString((9, 1), default="OML")
    message_code = HprimString((9, 2), default="O33")
    control_id = HprimString(10)
    processing_id = HprimString(11, default="P")
    version_id = HprimString(12, default="2.5")
    accept_acknowledgment_type = HprimString(15, default="AL")
    application_acknowledgment = HprimString(16, default="NE")

    patients = HprimForeign("PID", "header")
    specimens = HprimForeign("SPM", "header")
    infos_samples = HprimForeign("SAC", "patient")
    orders = HprimForeign("OBR", "header")


class AmsSegmentPID(Segment):
    patient_id = HprimString(4)
    patient_surname = HprimString((6, 1))
    patient_name = HprimString((6, 2))
    mother_maiden_name = HprimString(7)
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    sex = HprimString(9)
    patient_address = HprimString(12)
    height = HprimString(13)
    weight = HprimString(14)

    visits = HprimForeign("PV1", "patient")
    orders = HprimForeign("OBR", "patient")
    comments = HprimForeign("NTE", "patient")
    specimens = HprimForeign("SPM", "patient")


class AmsSegmentPV1(Segment):
    assigned_patient_location = HprimString(4)
    admission_date = HprimDate((5, 1))
    admission_status = HprimString((5, 2))
    attending_doctor = HprimString(8)
    hospital_service = HprimString((11, 1))
    hospital_institution = HprimString((11, 2))
    patient_type = HprimString(19)
    diet_type = HprimString(39)
    medication = HprimString(40)
    dosage_category = HprimString(41)

    orders = HprimForeign("OBR", "patient")
    specimens = HprimForeign("SPM", "patient")


class AmsSegmentSPM(Segment):
    # Sample ID: this is the univocal Sample Barcode Number
    specimen_id = HprimString(3)
    # Primary tube code
    specimen_parent_id = HprimString(4)
    specimen_type = HprimString(5)

    orders = HprimForeign("OBR", "patient")
    comments = HprimForeign("NTE", "patient")
    infos_samples = HprimForeign("SAC", "patient")
    orc = HprimForeign("ORC", "patient")


class AmsSegmentORC(Segment):
    request_control = HprimString(2)
    placer_order_number = HprimString(3)
    filler_order_number = HprimString(4)
    sample_dilution = HprimString(8)
    ward_phone_number = HprimString((23, 1))
    physician_phone_number = HprimString((23, 2))
    reserved = HprimString(24, necessary=True)

    orders = HprimForeign("OBR", "order")
    results = HprimForeign("OBX", "order")


class AmsSegmentSAC(Segment):
    request_number = HprimString(3)
    sample_id = HprimString(4)
    carrier_identifier = HprimString(11)
    position = HprimString(12)
    rack_location = HprimString((16, 1))
    bay_number = HprimString((16, 2))

    orders = HprimForeign("OBR", "patient")
    common_order = HprimForeign("ORC", "patient")
    specimens = HprimForeign("SPM", "patient")


class AmsSegmentNTE(Segment):
    comment_source = HprimString(3)
    comment_text = HprimString(4)


class AmsSegmentOBR(Segment):
    placer_order_number = HprimString(3)
    filler_order_number = HprimString(4)
    test_id = HprimString(5)
    collector_id = HprimString(11)
    clinical_data_code = HprimString((14, 1))
    clinical_data_type = HprimString((14, 2))
    request_date = HprimDate(15)
    requesting_provider = HprimString(17)
    request_laboratory = HprimString(25)
    quantity = HprimString((29, 1))
    sample_draw_date = HprimDate((29, 4))
    priority = HprimString((29, 6))

    results = HprimForeign("OBX", "order")
    infos_tests = HprimForeign("TCD", "order")
    comments = HprimForeign("NTE", "order")


class AmsSegmentDG1(Segment):
    nature_diagnostic_code = HprimString(8)
    diagnostic_code = HprimString(9)


class AmsSegmentOBX(Segment):
    value_type = HprimString(3)
    test_id = HprimString(4)
    target_name = HprimString(5)
    value = HprimString(6)
    units = HprimString(7)
    reference_ranges = HprimString(8)
    abnormal_flag = HprimString(9)
    instrument_flag = HprimString(11)
    result_status = HprimString(12)
    date_result = HprimDate(15)
    validation_user_ams = HprimString((17, 1))
    validation_user_instrument = HprimString((17, 2))
    operator_logged = HprimString((17, 3))
    second_validation_user_ams = HprimString((17, 4))
    instrument_code = HprimString((19, 1))
    instrument = HprimString((19, 2))
    path_id = HprimString((19, 3))
    lane_id = HprimString((19, 4))
    result_date_validation = HprimDate((20, 1))
    result_date_instrument = HprimDate((20, 2))
    result_date_second_validation = HprimDate((20, 3))

    inventory = HprimForeign("INV", "result")
    comments = HprimForeign("NTE", "result")
    infos_tests = HprimForeign("TCD", "result")


class AmsSegmentTCD(Segment):
    test_id = HprimString(2)
    dilution = HprimString(3)


class AmsSegmentINV(Segment):
    substance_id = HprimString(2)
    substance_identifier = HprimString((4, 1))
    substance_text = HprimString((4, 2))
    substance_reference_hl7 = HprimString((4, 3))
    serial_number_reagent = HprimString(5)
    expiration_date = HprimDate(13)
    first_used_date = HprimDate(14)
    lot_reagent = HprimString(17)


class Driver(AstmMessageV2):
    SPEC = {"code": "ams", "label": "AMS HL7 ABBOTT"}
    SEGMENT_CLASS_MAP = {
        "MSH": AmsSegmentMSH,
        "PID": AmsSegmentPID,
        "PV1": AmsSegmentPV1,
        "SPM": AmsSegmentSPM,
        "SAC": AmsSegmentSAC,
        "OBR": AmsSegmentOBR,
        "OBX": AmsSegmentOBX,
        "NTE": AmsSegmentNTE,
        "ORC": AmsSegmentORC,
        "DG1": AmsSegmentDG1,
        "INV": AmsSegmentINV,
        "TCD": AmsSegmentTCD,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"
    OPTIONS = [
        AstmMessageV2.Option(
            "ams_sender_name", label="Code émetteur (MSH-3.1)", default="MCA"
        ),
        AstmMessageV2.BooleanOption(
            "ams_exclure_preliminaire",
            label="Ignorer les résultats préliminaires (P) champ OBX-11",
            default=False,
        ),
    ]

    def __init__(self, **kwargs):
        self.sender_name = kwargs.pop("ams_sender_name", "MCA")
        self.exclure_preliminaire = kwargs.pop("ams_exclure_preliminaire", False)
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        return AmsSegmentMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def encode_worklist(self, worklist, context):
        msg = self.create(
            sender_name=self.sender_name,
            date=context["now"],
            control_id=context.get("control_id", int(time.time() * 1000)),
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority", "rerun"],
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
            action={"del": "CA", "add": "AT", "new": "NW"},
        ):
            seg_pid = msg.patients.add(
                patient_id=patient.patient_id,
                patient_surname=patient.last_name,
                patient_name=patient.first_name,
                mother_maiden_name=patient.maiden_name,
                birthdate=patient.birthdate,
                height=patient.height,
                weight=patient.weight,
                sex=patient.sex,
            )
            physician = [p for p in patient.physicians if p.type.lower() == "attending"]
            if patient.location:
                seg_pv1 = seg_pid.visits.add(
                    attending_doctor=physician[0].code if physician else "",
                    hospital_service=patient.location,
                    patient_type="0" if patient.location else "1",
                )
            else:
                seg_pv1 = seg_pid
            for sample, action, priority, rerun in patient.samples:
                seg_spm = seg_pv1.specimens.add(
                    specimen_id=sample.sample_id, specimen_type=sample.specimen
                )
                for order in sample["orders"]:
                    if order.action == "add" and rerun:
                        orc = seg_spm.orc.add(
                            request_control="RR",
                            filler_order_number=sample.sample_id,
                            sample_dilution=order.dilution
                            if order.dilution
                            else "UNDILUTED",
                        )
                    else:
                        orc = seg_spm.orc.add(
                            request_control=action,
                            filler_order_number=sample.sample_id,
                            sample_dilution=order.dilution,
                        )
                    seg_obr = orc.orders.add(
                        test_id=order.code,
                        request_date=sample.request_date,
                        priority=priority,
                        requesting_provider=sample.lab,
                    )
                    if order.dilution:
                        seg_obr.comments.add(
                            comment_source="L",
                            comment_text=order.dilution
                            if order.dilution
                            else "UNDILUTED",
                        )
        if len(msg.patients):
            return msg

    def decode_message(self, message):
        worklist = Worklist(context="result")
        for seg_pid in message.patients:
            if seg_pid.visits:
                seg_pv1 = seg_pid.visits[0]
            else:
                seg_pv1 = seg_pid
            for seg_spm in seg_pv1.specimens:
                if seg_pid.patient_id == "QC":
                    sample = worklist.get_or_create(
                        sample_id="{}-{}".format(
                            seg_pid.patient_name, seg_spm.specimen_id
                        ),
                        control_lot="{}-{}".format(
                            seg_pid.patient_name, seg_spm.specimen_id
                        ),
                        type="control",
                    )
                else:
                    sample = worklist.get_or_create(
                        sample_id=seg_spm.specimen_id, type="patient"
                    )
                for seg_sac in seg_spm.infos_samples:
                    if seg_sac.carrier_identifier and seg_sac.position:
                        sample.rack = "{}-{}".format(
                            seg_sac.carrier_identifier, seg_sac.position
                        )
                    for seg_obr in seg_sac.orders:
                        for seg_obx in seg_obr.results:
                            if (
                                self.exclure_preliminaire
                                and seg_obx.result_status == "P"
                            ):
                                continue
                            comments = []
                            for seg_nte in seg_obx.comments:
                                comments.append(seg_nte.comment_text)
                            dilution = ""
                            for seg_tcd in seg_obx.infos_tests:
                                if seg_tcd.dilution != "UNDILUTED":
                                    dilution = seg_tcd.dilution
                            result = sample.results.add(
                                code=seg_obx.test_id,
                                value=seg_obx.value,
                                unit=seg_obx.units,
                                date=seg_obx.result_date_instrument,
                                flags=[seg_obx.instrument_flag],
                                dilution=dilution,
                                comments=comments,
                                operator=seg_obx.validation_user_instrument
                                if seg_obx.validation_user_instrument
                                else seg_obx.validation_user_ams,
                                instrument=seg_obx.instrument_code,
                            )
                            for seg_inv in seg_obx.inventory:
                                result.lots.add(
                                    type="reagent", lot=seg_inv.serial_number_reagent
                                )

        return worklist
